@@ -1,36 +1,11 @@
 /*
- * "$Id: image.cxx 1708 2013-08-09 02:53:39Z msweet $"
+ * Image handling routines for HTMLDOC, a HTML document processing program.
  *
- *   Image handling routines for HTMLDOC, a HTML document processing program.
+ * Copyright © 2011-2021 by Michael R Sweet.
+ * Copyright © 1997-2010 by Easy Software Products.  All rights reserved.
  *
- *   Copyright 2011 by Michael R Sweet.
- *   Copyright 1997-2010 by Easy Software Products.  All rights reserved.
- *
- *   This program is free software.  Distribution and use rights are outlined in
- *   the file "COPYING.txt".
- *
- * Contents:
- *
- *   gif_read_cmap()      - Read the colormap from a GIF file...
- *   gif_get_block()      - Read a GIF data block...
- *   gif_get_code()       - Get a LZW code from the file...
- *   gif_read_image()     - Read a GIF image stream...
- *   gif_read_lzw()       - Read a byte from the LZW stream...
- *   image_compare()      - Compare two image filenames...
- *   image_copy()         - Copy image files to the destination directory...
- *   image_flush_cache()  - Flush the image cache...
- *   image_getlist()      - Get the list of images that are loaded.
- *   image_load()         - Load an image file from disk...
- *   image_load_bmp()     - Read a BMP image file.
- *   image_load_gif()     - Load a GIF image file...
- *   image_load_jpeg()    - Load a JPEG image file.
- *   image_load_png()     - Load a PNG image file.
- *   image_need_mask()    - Allocate memory for the image mask...
- *   image_set_mask()     - Set a bit in the image mask.
- *   jpeg_error_handler() - Handle JPEG errors by not exiting.
- *   read_word()          - Read a 16-bit unsigned integer.
- *   read_dword()         - Read a 32-bit unsigned integer.
- *   read_long()          - Read a 32-bit signed integer.
+ * This program is free software.  Distribution and use rights are outlined in
+ * the file "COPYING".
  */
 
 /*
@@ -39,11 +14,15 @@
 
 #include "htmldoc.h"
 
+#ifdef HAVE_LIBJPEG
 extern "C" {		/* Workaround for JPEG header problems... */
-#include <jpeglib.h>	/* JPEG/JFIF image definitions */
+#  include <jpeglib.h>	/* JPEG/JFIF image definitions */
 }
+#endif // HAVE_JPEG
 
-#include <png.h>	/* Portable Network Graphics (PNG) definitions */
+#ifdef HAVE_LIBPNG
+#  include <png.h>	/* Portable Network Graphics (PNG) definitions */
+#endif // HAVE_LIBPNG
 
 
 /*
@@ -72,7 +51,7 @@ typedef uchar	gif_cmap_t[256][3];
  * Local globals...
  */
 
-static int	num_images = 0,		/* Number of images in cache */
+static size_t	num_images = 0,		/* Number of images in cache */
 		alloc_images = 0;	/* Allocated images */
 static image_t	**images = NULL;	/* Images in cache */
 static int	gif_eof = 0;		/* Did we hit EOF? */
@@ -93,12 +72,19 @@ static int	gif_read_lzw(FILE *fp, int first_time, int input_code_size);
 static int	image_compare(image_t **img1, image_t **img2);
 static int	image_load_bmp(image_t *img, FILE *fp, int gray, int load_data);
 static int	image_load_gif(image_t *img, FILE *fp, int gray, int load_data);
+
+#ifdef HAVE_LIBJPEG
 static int	image_load_jpeg(image_t *img, FILE *fp, int gray, int load_data);
+static void	jpeg_error_handler(j_common_ptr);
+#endif // HAVE_LIBJPEG
+
+#ifdef HAVE_LIBPNG
 static int	image_load_png(image_t *img, FILE *fp, int gray, int load_data);
+#endif // HAVE_LIBPNG
+
 static void	image_need_mask(image_t *img, int scaling = 1);
 static void	image_set_mask(image_t *img, int x, int y, uchar alpha = 0);
 
-static void		jpeg_error_handler(j_common_ptr);
 static int		read_long(FILE *fp);
 static unsigned short	read_word(FILE *fp);
 static unsigned int	read_dword(FILE *fp);
@@ -121,7 +107,7 @@ gif_read_cmap(FILE       *fp,		/* I  - File to read from */
   * Read the colormap...
   */
 
-  if (fread(cmap, 3, ncolors, fp) < (size_t)ncolors)
+  if (fread(cmap, 3, (size_t)ncolors, fp) < (size_t)ncolors)
   {
     progress_error(HD_ERROR_READ_ERROR,
                    "Unable to read GIF colormap: %s", strerror(errno));
@@ -177,7 +163,7 @@ gif_get_block(FILE  *fp,	/* I - File to read from */
   }
   else if (count == 0)
     gif_eof = 1;
-  else if (fread(buf, 1, count, fp) < (size_t)count)
+  else if (fread(buf, 1, (size_t)count, fp) < (size_t)count)
   {
     progress_error(HD_ERROR_READ_ERROR,
                    "Unable to read GIF block of %d bytes: %s", count,
@@ -230,7 +216,7 @@ gif_get_code(FILE *fp,		/* I - File to read from */
     return (0);
   }
 
-  if ((curbit + code_size) >= lastbit)
+  if ((curbit + (unsigned)code_size) >= lastbit)
   {
    /*
     * Don't have enough bits to hold the code...
@@ -278,19 +264,18 @@ gif_get_code(FILE *fp,		/* I - File to read from */
     */
 
     curbit    = (curbit - lastbit) + 8 * last_byte;
-    last_byte += count;
+    last_byte += (unsigned)count;
     lastbit   = last_byte * 8;
   }
 
-  ret = 0;
-  for (ret = 0, i = curbit + code_size - 1, j = code_size;
+  for (ret = 0, i = curbit + (unsigned)code_size - 1, j = (unsigned)code_size;
        j > 0;
        i --, j --)
     ret = (ret << 1) | ((buf[i / 8] & bits[i & 7]) != 0);
 
-  curbit += code_size;
+  curbit += (unsigned)code_size;
 
-  return ret;
+  return (int)ret;
 }
 
 
@@ -318,7 +303,7 @@ gif_read_image(FILE       *fp,		/* I - Input file */
   xpos      = 0;
   ypos      = 0;
   pass      = 0;
-  code_size = getc(fp);
+  code_size = (uchar)getc(fp);
 
   if (gif_read_lzw(fp, 1, code_size) < 0)
     return (-1);
@@ -382,17 +367,17 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
 		code,			/* Current code */
 		incode;			/* Input code */
   static short	fresh = 0,		/* 1 = empty buffers */
-		code_size,		/* Current code size */
-		set_code_size,		/* Initial code size set */
-		max_code,		/* Maximum code used */
-		max_code_size,		/* Maximum code size */
-		firstcode,		/* First code read */
-		oldcode,		/* Last code read */
-		clear_code,		/* Clear code for LZW input */
-		end_code,		/* End code for LZW input */
+		code_size = 0,		/* Current code size */
+		set_code_size = 0,	/* Initial code size set */
+		max_code = 0,		/* Maximum code used */
+		max_code_size = 0,	/* Maximum code size */
+		firstcode = 0,		/* First code read */
+		oldcode = 0,		/* Last code read */
+		clear_code = 0,		/* Clear code for LZW input */
+		end_code = 0,		/* End code for LZW input */
 		table[2][4096],		/* String table */
 		stack[8192],		/* Output stack */
-		*sp;			/* Current stack pointer */
+		*sp = stack;		/* Current stack pointer */
 
 
   if (first_time)
@@ -401,9 +386,9 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
     * Setup LZW state...
     */
 
-    set_code_size = input_code_size;
+    set_code_size = (short)input_code_size;
     code_size     = set_code_size + 1;
-    clear_code    = 1 << set_code_size;
+    clear_code    = (short)(1 << set_code_size);
     end_code      = clear_code + 1;
     max_code_size = 2 * clear_code;
     max_code      = clear_code + 2;
@@ -423,7 +408,7 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
     for (i = 0; i < clear_code; i ++)
     {
       table[0][i] = 0;
-      table[1][i] = i;
+      table[1][i] = (short)i;
     }
 
     for (; i < 4096; i ++)
@@ -438,7 +423,7 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
     fresh = 0;
 
     do
-      firstcode = oldcode = gif_get_code(fp, code_size, 0);
+      firstcode = oldcode = (short)gif_get_code(fp, code_size, 0);
     while (firstcode == clear_code);
 
     return (firstcode);
@@ -454,7 +439,7 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
       for (i = 0; i < clear_code; i ++)
       {
 	table[0][i] = 0;
-	table[1][i] = i;
+	table[1][i] = (short)i;
       }
 
       for (; i < 4096; i ++)
@@ -466,7 +451,7 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
 
       sp = stack;
 
-      firstcode = oldcode = gif_get_code(fp, code_size, 0);
+      firstcode = oldcode = (short)gif_get_code(fp, code_size, 0);
 
       return (firstcode);
     }
@@ -514,7 +499,7 @@ gif_read_lzw(FILE *fp,			/* I - File to read from */
       }
     }
 
-    oldcode = incode;
+    oldcode = (short)incode;
 
     if (sp > stack)
       return (*--sp);
@@ -590,7 +575,7 @@ image_copy(const char *src,		/* I - Source file */
   }
 
   while ((nbytes = fread(buffer, 1, sizeof(buffer), in)) > 0)
-    fwrite(buffer, 1, nbytes, out);
+    fwrite(buffer, 1, (size_t)nbytes, out);
 
   progress_error(HD_ERROR_NONE, "BYTES: %ld", ftell(out));
 
@@ -631,7 +616,7 @@ image_find(const char *filename,/* I - Name of image file */
     strlcpy(key.filename, filename, sizeof(key.filename));
     keyptr = &key;
 
-    match = (image_t **)bsearch(&keyptr, images, num_images, sizeof(image_t *),
+    match = (image_t **)bsearch(&keyptr, images, (size_t)num_images, sizeof(image_t *),
                                 (int (*)(const void *, const void *))image_compare);
     if (match != NULL)
     {
@@ -653,7 +638,7 @@ image_find(const char *filename,/* I - Name of image file */
 void
 image_flush_cache(void)
 {
-  int	i;			/* Looping var */
+  size_t	i;			/* Looping var */
 
 
  /*
@@ -740,7 +725,7 @@ image_load(const char *filename,/* I - Name of image file */
     strlcpy(key.filename, filename, sizeof(key.filename));
     keyptr = &key;
 
-    match = (image_t **)bsearch(&keyptr, images, num_images, sizeof(image_t *),
+    match = (image_t **)bsearch(&keyptr, images, (size_t)num_images, sizeof(image_t *),
                                 (int (*)(const void *, const void *))image_compare);
     if (match != NULL && (!load_data || (*match)->pixels))
     {
@@ -781,7 +766,7 @@ image_load(const char *filename,/* I - Name of image file */
 #ifdef DEBUG
   printf("Header for \"%s\" (%s): \"", filename, realname);
 
-  for (i = 0; i < sizeof(header); i ++)
+  for (i = 0; i < (int)sizeof(header); i ++)
     if (header[i] < ' ' || header[i] >= 127)
       printf("\\x%02X", header[i]);
     else
@@ -789,7 +774,7 @@ image_load(const char *filename,/* I - Name of image file */
 
   puts("\"\n");
 
-  printf("match = %p\n", match);
+  printf("match = %p\n", (void *)match);
 #endif // DEBUG
 
   rewind(fp);
@@ -811,7 +796,7 @@ image_load(const char *filename,/* I - Name of image file */
       {
 	progress_error(HD_ERROR_OUT_OF_MEMORY,
 	               "Unable to allocate memory for %d images - %s",
-                       alloc_images, strerror(errno));
+                       (int)alloc_images, strerror(errno));
 	fclose(fp);
 	return (NULL);
       }
@@ -844,13 +829,17 @@ image_load(const char *filename,/* I - Name of image file */
     status = image_load_gif(img,  fp, gray, load_data);
   else if (memcmp(header, "BM", 2) == 0)
     status = image_load_bmp(img, fp, gray, load_data);
+#ifdef HAVE_LIBPNG
   else if (memcmp(header, "\211PNG", 4) == 0)
     status = image_load_png(img, fp, gray, load_data);
+#endif // HAVE_LIBPNG
+#ifdef HAVE_LIBJPEG
   else if (memcmp(header, "\377\330\377", 3) == 0)
     status = image_load_jpeg(img, fp, gray, load_data);
+#endif // HAVE_LIBJPEG
   else
   {
-    progress_error(HD_ERROR_BAD_FORMAT, "Unknown image file format for \"%s\"!",
+    progress_error(HD_ERROR_BAD_FORMAT, "Unknown image file format for \"%s\".",
                    file_rlookup(filename));
     fclose(fp);
     free(img);
@@ -914,16 +903,16 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
   read_dword(fp);
 
   // Then the bitmap information...
-  info_size        = read_dword(fp);
+  info_size        = (int)read_dword(fp);
   img->width       = read_long(fp);
   img->height      = read_long(fp);
   read_word(fp);
   depth            = read_word(fp);
-  compression      = read_dword(fp);
+  compression      = (int)read_dword(fp);
   read_dword(fp);
   read_long(fp);
   read_long(fp);
-  colors_used      = read_dword(fp);
+  colors_used      = (int)read_dword(fp);
   read_dword(fp);
 
   if (info_size > 40)
@@ -934,7 +923,7 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
   if (colors_used == 0 && depth <= 8)
     colors_used = 1 << depth;
 
-  fread(colormap, colors_used, 4, fp);
+  fread(colormap, (size_t)colors_used, 4, fp);
 
   // Setup image and buffers...
   img->depth  = gray ? 1 : 3;
@@ -948,7 +937,7 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
   if (!load_data)
     return (0);
 
-  img->pixels = (uchar *)malloc(img->width * img->height * img->depth);
+  img->pixels = (uchar *)malloc((size_t)(img->width * img->height * img->depth));
   if (img->pixels == NULL)
     return (-1);
 
@@ -978,7 +967,7 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
           for (x = img->width, bit = 128; x > 0; x --)
 	  {
 	    if (bit == 128)
-	      byte = getc(fp);
+	      byte = (uchar)getc(fp);
 
 	    if (byte & bit)
 	    {
@@ -1218,16 +1207,16 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
 	      temp = getc(fp) * 8;
 	      temp += getc(fp) * 61;
 	      temp += getc(fp) * 31;
-	      *ptr++ = temp / 100;
+	      *ptr++ = (uchar)(temp / 100);
 	    }
 	  }
 	  else
 	  {
             for (x = img->width; x > 0; x --, ptr += 3)
 	    {
-	      ptr[2] = getc(fp);
-	      ptr[1] = getc(fp);
-	      ptr[0] = getc(fp);
+	      ptr[2] = (uchar)getc(fp);
+	      ptr[1] = (uchar)getc(fp);
+	      ptr[0] = (uchar)getc(fp);
 	    }
           }
 
@@ -1271,6 +1260,9 @@ image_load_gif(image_t *img,	/* I - Image pointer */
   img->height = (buf[9] << 8) | buf[8];
   ncolors     = 2 << (buf[10] & 0x07);
 
+  if (img->width <= 0 || img->width > 32767 || img->height <= 0 || img->height > 32767)
+    return (-1);
+
   // If we are writing an encrypted PDF file, bump the use count so we create
   // an image object (Acrobat 6 bug workaround)
   if (Encryption)
@@ -1290,7 +1282,7 @@ image_load_gif(image_t *img,	/* I - Image pointer */
           return (-1);		/* Early end of file */
 
       case '!' :	/* Extension record */
-          buf[0] = getc(fp);
+          buf[0] = (uchar)getc(fp);
           if (buf[0] == 0xf9)	/* Graphic Control Extension */
           {
             gif_get_block(fp, buf);
@@ -1349,7 +1341,7 @@ image_load_gif(image_t *img,	/* I - Image pointer */
 	  if (!load_data)
 	    return (0);
 
-          img->pixels = (uchar *)malloc(img->width * img->height * img->depth);
+          img->pixels = (uchar *)malloc((size_t)(img->width * img->height * img->depth));
           if (img->pixels == NULL)
             return (-1);
 
@@ -1359,6 +1351,16 @@ image_load_gif(image_t *img,	/* I - Image pointer */
 }
 
 
+#ifdef HAVE_LIBJPEG
+typedef struct hd_jpeg_err_s	// JPEG error manager extension
+{
+  struct jpeg_error_mgr	jerr;	// JPEG error manager information
+  jmp_buf	retbuf;		// setjmp() return buffer
+  char		message[JMSG_LENGTH_MAX];
+				// Last error message
+} hd_jpeg_err_t;
+
+
 /*
  * 'image_load_jpeg()' - Load a JPEG image file.
  */
@@ -1370,19 +1372,26 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
                 int     load_data)/* I - 1 = load image data, 0 = just info */
 {
   struct jpeg_decompress_struct	cinfo;		/* Decompressor info */
-  struct jpeg_error_mgr		jerr;		/* Error handler info */
-  JSAMPROW			row;		/* Sample row pointer */
+  hd_jpeg_err_t			jerr;		// JPEG error handler
+JSAMPROW			row;		/* Sample row pointer */
 
 
-  jpeg_std_error(&jerr);
-  jerr.error_exit = jpeg_error_handler;
+  jpeg_std_error(&jerr.jerr);
+  jerr.jerr.error_exit = jpeg_error_handler;
 
-  cinfo.err = &jerr;
+  if (setjmp(jerr.retbuf))
+  {
+    progress_error(HD_ERROR_BAD_FORMAT, "%s (%s)", jerr.message,  file_rlookup(img->filename));
+    jpeg_destroy_decompress(&cinfo);
+    return (-1);
+  }
+
+  cinfo.err = (struct jpeg_error_mgr *)&jerr;
   jpeg_create_decompress(&cinfo);
   jpeg_stdio_src(&cinfo, fp);
   jpeg_read_header(&cinfo, (boolean)1);
 
-  cinfo.quantize_colors = 0;
+  cinfo.quantize_colors = FALSE;
 
   if (gray || cinfo.num_components == 1)
   {
@@ -1408,9 +1417,9 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
 
   jpeg_calc_output_dimensions(&cinfo);
 
-  img->width  = cinfo.output_width;
-  img->height = cinfo.output_height;
-  img->depth  = cinfo.output_components;
+  img->width  = (int)cinfo.output_width;
+  img->height = (int)cinfo.output_height;
+  img->depth  = (int)cinfo.output_components;
 
   if (!load_data)
   {
@@ -1418,7 +1427,7 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
     return (0);
   }
 
-  img->pixels = (uchar *)malloc(img->width * img->height * img->depth);
+  img->pixels = (uchar *)malloc((size_t)(img->width * img->height * img->depth));
 
   if (img->pixels == NULL)
   {
@@ -1430,9 +1439,7 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
 
   while (cinfo.output_scanline < cinfo.output_height)
   {
-    row = (JSAMPROW)(img->pixels +
-                     cinfo.output_scanline * cinfo.output_width *
-                     cinfo.output_components);
+    row = (JSAMPROW)(img->pixels + (size_t)cinfo.output_scanline * (size_t)cinfo.output_width * (size_t)cinfo.output_components);
     jpeg_read_scanlines(&cinfo, &row, (JDIMENSION)1);
   }
 
@@ -1441,8 +1448,10 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
 
   return (0);
 }
+#endif // HAVE_LIBJPEG
 
 
+#ifdef HAVE_LIBPNG
 /*
  * 'image_load_png()' - Load a PNG image file.
  */
@@ -1460,6 +1469,8 @@ image_load_png(image_t *img,	/* I - Image pointer */
   png_bytep	*rows;		/* PNG row pointers */
   uchar		*inptr,		/* Input pixels */
 		*outptr;	/* Output pixels */
+  int		color_type,	/* PNG color mode */
+		bit_depth;	/* PNG bit depth */
 
 
  /*
@@ -1508,30 +1519,49 @@ image_load_png(image_t *img,	/* I - Image pointer */
 
   png_init_io(pp, fp);
 
+#  if defined(PNG_SKIP_sRGB_CHECK_PROFILE) && defined(PNG_SET_OPTION_SUPPORTED)
+  // Don't throw errors with "invalid" sRGB profiles produced by Adobe apps.
+  png_set_option(pp, PNG_SKIP_sRGB_CHECK_PROFILE, PNG_OPTION_ON);
+#  endif // PNG_SKIP_sRGB_CHECK_PROFILE && PNG_SET_OPTION_SUPPORTED
+
  /*
   * Get the image dimensions and convert to grayscale or RGB...
   */
 
   png_read_info(pp, info);
 
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_PALETTE)
+  bit_depth  = png_get_bit_depth(pp, info);
+  color_type = png_get_color_type(pp, info);
+
+  if (png_get_valid(pp, info, PNG_INFO_tRNS))
   {
-    png_set_expand(pp);
+    png_set_tRNS_to_alpha(pp);
+    color_type |= PNG_COLOR_MASK_ALPHA;
+  }
+
+  if (color_type & PNG_COLOR_MASK_PALETTE)
+  {
+    png_set_palette_to_rgb(pp);
 
     // If we are writing an encrypted PDF file, bump the use count so we create
     // an image object (Acrobat 6 bug workaround)
     if (Encryption)
       img->use ++;
   }
-  else if (png_get_bit_depth(pp, info) < 8)
+  else if (!(color_type & PNG_COLOR_MASK_COLOR) && bit_depth < 8)
   {
-    png_set_packing(pp);
-    png_set_expand(pp);
+    png_set_expand_gray_1_2_4_to_8(pp);
   }
-  else if (png_get_bit_depth(pp, info) == 16)
+  else if (bit_depth == 16)
+  {
+#  if PNG_LIBPNG_VER >= 10504
+    png_set_scale_16(pp);
+#  else
     png_set_strip_16(pp);
+#  endif // PNG_LIBPNG_VER >= 10504
+  }
 
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_COLOR)
+  if (color_type & PNG_COLOR_MASK_COLOR)
   {
     depth      = 3;
     img->depth = gray ? 1 : 3;
@@ -1542,10 +1572,10 @@ image_load_png(image_t *img,	/* I - Image pointer */
     img->depth = 1;
   }
 
-  img->width  = png_get_image_width(pp, info);
-  img->height = png_get_image_height(pp, info);
+  img->width  = (int)png_get_image_width(pp, info);
+  img->height = (int)png_get_image_height(pp, info);
 
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_ALPHA)
+  if (color_type & PNG_COLOR_MASK_ALPHA)
   {
     if ((PSLevel == 0 && PDFVersion >= 14) || PSLevel == 3)
       image_need_mask(img, 8);
@@ -1557,18 +1587,17 @@ image_load_png(image_t *img,	/* I - Image pointer */
     depth ++;
   }
 
-#ifdef DEBUG
-  printf("color_type=0x%04x, depth=%d, img->width=%d, img->height=%d, img->depth=%d\n",
-         png_get_color_type(pp, info), depth, img->width, img->height, img->depth);
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_COLOR)
+#  ifdef DEBUG
+  printf("bit_depth=%d, color_type=0x%04x, depth=%d, img->width=%d, img->height=%d, img->depth=%d\n", bit_depth, color_type, depth, img->width, img->height, img->depth);
+  if (color_type & PNG_COLOR_MASK_COLOR)
     puts("    COLOR");
   else
     puts("    GRAYSCALE");
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_ALPHA)
+  if (color_type & PNG_COLOR_MASK_ALPHA)
     puts("    ALPHA");
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_PALETTE)
+  if (color_type & PNG_COLOR_MASK_PALETTE)
     puts("    PALETTE");
-#endif // DEBUG
+#  endif // DEBUG
 
   if (!load_data)
   {
@@ -1576,7 +1605,7 @@ image_load_png(image_t *img,	/* I - Image pointer */
     return (0);
   }
 
-  img->pixels = (uchar *)malloc(img->width * img->height * depth);
+  img->pixels = (uchar *)calloc(1,(size_t)(img->width * img->height * depth));
 
  /*
   * Allocate pointers...
@@ -1592,15 +1621,15 @@ image_load_png(image_t *img,	/* I - Image pointer */
   */
 
   for (i = png_set_interlace_handling(pp); i > 0; i --)
-    png_read_rows(pp, rows, NULL, img->height);
+    png_read_rows(pp, rows, NULL, (png_uint_32)img->height);
 
  /*
   * Generate the alpha mask as necessary...
   */
 
-  if (png_get_color_type(pp, info) & PNG_COLOR_MASK_ALPHA)
+  if (color_type & PNG_COLOR_MASK_ALPHA)
   {
-#ifdef DEBUG
+#  ifdef DEBUG
     for (inptr = img->pixels, i = 0; i < img->height; i ++)
     {
       for (j = 0; j < img->width; j ++, inptr += depth)
@@ -1616,7 +1645,7 @@ image_load_png(image_t *img,	/* I - Image pointer */
 
       putchar('\n');
     }
-#endif // DEBUG
+#  endif // DEBUG
 
     for (inptr = img->pixels + depth - 1, i = 0; i < img->height; i ++)
       for (j = 0; j < img->width; j ++, inptr += depth)
@@ -1627,10 +1656,10 @@ image_load_png(image_t *img,	/* I - Image pointer */
   * Reformat the data as necessary for the reader...
   */
 
-  if (gray && png_get_color_type(pp, info) & PNG_COLOR_MASK_COLOR)
+  if (gray && (color_type & PNG_COLOR_MASK_COLOR))
   {
    /*
-    * Greyscale output needed...
+    * Grayscale output needed...
     */
 
     for (inptr = img->pixels, outptr = img->pixels, i = img->width * img->height;
@@ -1675,6 +1704,7 @@ image_load_png(image_t *img,	/* I - Image pointer */
 
   return (0);
 }
+#endif // HAVE_LIBPNG
 
 
 /*
@@ -1682,10 +1712,10 @@ image_load_png(image_t *img,	/* I - Image pointer */
  */
 
 static void
-image_need_mask(image_t *img,	/* I - Image to add mask to */
-                int     scaling)/* I - Scaling for mask image */
+image_need_mask(image_t *img,		/* I - Image to add mask to */
+                int     scaling)	/* I - Scaling for mask image */
 {
-  int	size;			/* Byte size of mask image */
+  size_t	size;			/* Byte size of mask image */
 
 
   if (img == NULL || img->mask != NULL)
@@ -1702,13 +1732,13 @@ image_need_mask(image_t *img,	/* I - Image to add mask to */
   {
     // Alpha image
     img->maskwidth = img->width;
-    size           = img->width * img->height;
+    size           = (size_t)(img->width * img->height);
   }
   else
   {
     // Alpha mask
     img->maskwidth = (img->width * scaling + 7) / 8;
-    size           = img->maskwidth * img->height * scaling + 1;
+    size           = (size_t)(img->maskwidth * img->height * scaling + 1);
   }
 
   img->mask = (uchar *)calloc(size, 1);
@@ -1797,15 +1827,25 @@ image_unload(image_t *img)	// I - Image
 }
 
 
+#ifdef HAVE_LIBJPEG
 /*
  * 'jpeg_error_handler()' - Handle JPEG errors by not exiting.
  */
 
 static void
-jpeg_error_handler(j_common_ptr)
+jpeg_error_handler(j_common_ptr p)	// Common JPEG data
 {
-  return;
+  hd_jpeg_err_t	*jerr = (hd_jpeg_err_t *)p->err;
+					// JPEG error handler
+
+
+  // Save the error message in the string buffer...
+  (jerr->jerr.format_message)(p, jerr->message);
+
+  // Return to the point we called setjmp()...
+  longjmp(jerr->retbuf, 1);
 }
+#endif // HAVE_LIBJPEG
 
 
 /*
@@ -1817,10 +1857,10 @@ read_word(FILE *fp)       /* I - File to read from */
 {
   unsigned char b0, b1; /* Bytes from file */
 
-  b0 = getc(fp);
-  b1 = getc(fp);
+  b0 = (uchar)getc(fp);
+  b1 = (uchar)getc(fp);
 
-  return ((b1 << 8) | b0);
+  return (unsigned short)((b1 << 8) | b0);
 }
 
 
@@ -1833,12 +1873,12 @@ read_dword(FILE *fp)              /* I - File to read from */
 {
   unsigned char b0, b1, b2, b3; /* Bytes from file */
 
-  b0 = getc(fp);
-  b1 = getc(fp);
-  b2 = getc(fp);
-  b3 = getc(fp);
+  b0 = (uchar)getc(fp);
+  b1 = (uchar)getc(fp);
+  b2 = (uchar)getc(fp);
+  b3 = (uchar)getc(fp);
 
-  return ((((((b3 << 8) | b2) << 8) | b1) << 8) | b0);
+  return (unsigned)((((((b3 << 8) | b2) << 8) | b1) << 8) | b0);
 }
 
 
@@ -1851,15 +1891,10 @@ read_long(FILE *fp)               /* I - File to read from */
 {
   unsigned char b0, b1, b2, b3; /* Bytes from file */
 
-  b0 = getc(fp);
-  b1 = getc(fp);
-  b2 = getc(fp);
-  b3 = getc(fp);
+  b0 = (uchar)getc(fp);
+  b1 = (uchar)getc(fp);
+  b2 = (uchar)getc(fp);
+  b3 = (uchar)getc(fp);
 
   return ((int)(((((b3 << 8) | b2) << 8) | b1) << 8) | b0);
 }
-
-
-/*
- * End of "$Id: image.cxx 1708 2013-08-09 02:53:39Z msweet $".
- */
